import os

/**
 Wraps the SDKs TTLServiceClient to ease some short commings of the SDK regarding connection state to device and backend.
 */
class ConnectionService: NSObject, TTLServiceEventDelegate {
    private let serviceClient : TTLServiceClient
    // use weak hash map, so delegates are removed once they are no longer referenced.
    private var delegates = NSHashTable<ConnectionServiceDelegate>(options: NSPointerFunctions.Options.weakMemory)
    /// A list of LINK100 devices that are available for communication.
    private(set) var devices : Set<String> = Set.init()
    /// Backend connection status per device identifier
    private var backendConnected = Dictionary<String, Bool>()
    
    init(serviceClient : TTLServiceClient) {
        self.serviceClient = serviceClient
        super.init()
        serviceClient.addServiceDelegate(self)
        serviceClient.scanForDevices()
        
    }
    
    //
    // MARK: - SDK callbacks
    //
    
    func connected(to deviceIdentifier: String) -> Bool {
        return devices.contains(deviceIdentifier)
    }
    
    func serverConnected(for deviceIdentifier: String) -> Bool {
        return backendConnected[deviceIdentifier] == true
    }
    
    func onDeviceDidAppear(_ deviceIdentifier: String) {
        guard !connected(to: deviceIdentifier) else { return }
        devices = devices.union(Set(arrayLiteral: deviceIdentifier))
        nofifyDelegates { (delegate) in
            delegate.onDeviceConnected?(deviceIdentifier: deviceIdentifier)
            delegate.onDeviceListChanged?(devices: devices)
        }
    }
    
    func onDeviceDidDisappear(_ deviceIdentifier: String) {
        guard connected(to: deviceIdentifier) else { return }
        devices = devices.subtracting(Set(arrayLiteral: deviceIdentifier))
        nofifyDelegates { (delegate) in
            delegate.onDeviceDisconnected?(deviceIdentifier: deviceIdentifier)
            delegate.onDeviceListChanged?(devices: devices)
        }
    }
    
    func onServerConnectionSuccessful(_ deviceIdentifier: String) {
        // Callback called as result of TTLServiceClient.connect(...) call, when a connection was established successfully.
        backendConnected[deviceIdentifier] = true
        nofifyDelegates { (delegate) in
            delegate.onServerConnected?(deviceIdentifier: deviceIdentifier)
        }
    }
    
    func onServerConnectionFailed(_ deviceIdentifier: String, reason: TTLServiceServerConnectionFailedReason) {
        // Callback called as result of TTLServiceClient.connect(...) call,
        // when the connection to the server could not be established.
        backendConnected.removeValue(forKey: deviceIdentifier)
        var error = ServerDisconnectReason.connectionLost
        if (reason.error as NSError).code == TTL_SERVICE_ERROR_CODE_SERVER_CONNECTION_WRONG_CREDENTIALS
        {
            error = .wrongCredentials
        }
        else if (reason.error as NSError).code == TTL_SERVICE_ERROR_CODE_SERVER_CONNECTION_WRONG_DEVICE_ID
        {
            error = .wrongDevice
        }
        nofifyDelegates { (delegate) in
            delegate.onServerDisconnected?(deviceIdentifier: deviceIdentifier, reason: error)
        }
    }
    
    func onServiceState(_ serviceStates: [TTLServiceStateResponse]) {
        serviceStates
            .filter({$0.serverConnection != .connected})
            .filter({backendConnected[$0.deviceIdentifier] == true})
            .forEach({ (serviceState) in
                // Server disconnected event. Called when the server was disconnected without errors,
                // e.g. because of time out or disconnect call to the SDK.
                backendConnected.removeValue(forKey: serviceState.deviceIdentifier)
                nofifyDelegates(block: { (delegate) in
                    delegate.onServerDisconnected?(deviceIdentifier: serviceState.deviceIdentifier, reason: .connectionLost)
                })
            })
    }
    
    //
    // MARK: - Delegate handling
    //
    
    func addDelegate(delegate: ConnectionServiceDelegate) {
        delegates.add(delegate)
    }
    
    private func nofifyDelegates(block: (ConnectionServiceDelegate) -> ()) {
        let enumerator = delegates.objectEnumerator()
        
        while let nextDelegate = enumerator.nextObject() as? ConnectionServiceDelegate {
            block(nextDelegate)
        }
    }
}

@objc protocol ConnectionServiceDelegate{
    @objc optional func onServerConnected(deviceIdentifier: String)
    @objc optional func onServerDisconnected(deviceIdentifier: String, reason: ServerDisconnectReason)
    @objc optional func onDeviceConnected(deviceIdentifier: String)
    @objc optional func onDeviceDisconnected(deviceIdentifier: String)
    @objc optional func onDeviceListChanged(devices: Set<String>)
}
