import os

/**
 Simple helper class that takes care of handling state changes. This allows to implement a
 simplified state mashine.
 */
final class StateHandler<Delegate: StateHandlerDelegate> {
    /// The current state the StateHandler is in.
    private(set) var state: Delegate.State
    /// A optional deviceIdentifier. If set events only for this deviceIdentifier will be considered.
    private var deviceIdentifier: String? = nil
    /// The delegate that will be informed about a successful state change.
    weak var delegate : Delegate? = nil
    
    init(initialState : Delegate.State, deviceIdentifier: String? = nil) {
        self.state = initialState
        self.deviceIdentifier = deviceIdentifier
    }
    
    /**
     Helper that handles state transitions. Informs the delegate on successful state changes.
     
     - Parameter event: The event that triggers the state change (only for logging purposes)
     - Parameter deviceIdentifier: The device the event was triggered for (or `nil` if the event has no device context).
       The state change is skipped if the event is for another device then the one set for the StateHandler.
     - Parameter expectedStates: A list of expected transition start states. The state change is skipped if the current state
       is not contained in the list.
     - Parameter newState: the new state we want to change to.
     - Parameter onSuccess: Block that will be executed if we actually changed the state. Note that transition from one state
       to the same state will be notified as well.
     */
    @discardableResult func changeState(on event: String,
                                                 of deviceIdentifier: String? = nil,
                                                 from expectedStates: [Delegate.State],
                                                 to newState: Delegate.State,
                                                 onSuccess: (() ->())? = nil) -> Bool
    {
        guard validDevice(deviceIdentifier) && expectedStates.contains(self.state) else { return false }
        os_log("%s event %s: changing state from %s to %s", log: createTraceLog(), String(reflecting: Delegate.self), event, String(reflecting: state), String(reflecting: newState))
        state = newState
        onSuccess?()
        delegate?.onStateChanged(newState: newState)
        return true
    }
    
    func filter(deviceIdentifier: String) {
        self.deviceIdentifier = deviceIdentifier
    }
    
    private func validDevice(_ deviceIdentifier: String?) -> Bool
    {
        return deviceIdentifier == nil || self.deviceIdentifier == deviceIdentifier
    }
}

protocol StateHandlerDelegate: class {
    associatedtype State : Equatable
    
    func onStateChanged(newState: State)
}
