import os

/**
 Presenter handling the login and activation process on the login screen. The activation is the process which links an user
 to an LINK100 device.
 
 The presenter uses a simplified state mashine to handle the activation process. The state mashine for the activation looks
 as follows, with states being enclosed in brackets ( [state] ) and events depicted as arrows ( -event-> ):
 
     // start of activation process, establish backend connection via serviceClient.connect(...)
     [start] | [finished] -activate-> [connectingToBackend]
     // connection established, start the actual activation via serviceClient.startActivation(...)
     [connectingToBackend] -onServerConnected-> [activating]
     // activation finished, disconnect from backend if neccessary via serviceClient.disconnectFromServer(...)
     [activating] -onActivationSuccessful-> [finished]
     // activation failed due to some unknown reason
     [connectingToBackend] -onActivationFailed-> [finished]
 
 The presenter needs to handle additional error events. As the SDK works in a stateless fashion, it's not aware that an
 activation is in progress once it was triggered. If the device or server connection fails inbetween it will not call
 onActivationFailed. However, it will inform about device or server disconnection via alternative service callbacks.
 Thus these are the additional error events the presenter needs to handle:
 
     // server connection could no be established
     [connectingToBackend] -onServerDisconnected-> [finished]
     // server connection lost during activation
     [activating] -onServerDisconnected-> [finished]
     // device connection lost while establishing backend connection
     [connectingToBackend] -onDeviceDisconnected-> [finished]
     // device connection lost while activating
     [activating] -onDeviceDisconnected-> [finished]
 */
class LoginActivationPresenter: NSObject , TTLServiceEventDelegate, ConnectionServiceDelegate, StateHandlerDelegate {
    private let serviceClient : TTLServiceClient
    private let connectionService: ConnectionService
    private let userService: UserService
    
    /// handler that takes care of state transistions during activation process
    private var stateHandler : StateHandler<LoginActivationPresenter>
    /// device used for activation
    private var device: String?
    /// credentials used for activation
    private var credentials: Credentials?
    /// backend used for activation
    private var backend: BackendType?
    /// delegate which is informed about state changes in the presenter
    weak var delegate : LoginActivationPresenterDelegate?
    
    init(serviceClient : TTLServiceClient, connectionService: ConnectionService, userService: UserService) {
        self.serviceClient = serviceClient
        self.connectionService = connectionService
        self.userService = userService
        self.stateHandler = StateHandler(initialState: .start)
        super.init()
        self.stateHandler.delegate = self
        self.connectionService.addDelegate(delegate: self)
        self.serviceClient.addServiceDelegate(self)

    }
    
    func activate(userName: String, password: String, device: String, prod: Bool) {
        stateHandler.changeState(on: "activate", from: [.start, .finished(error: nil)] , to: .connectingToBackend)
        {
            self.stateHandler.filter(deviceIdentifier: device)
            let credentials = Credentials(account: userName, password: password)
            let backend : BackendType = prod ? .Prod : .Dev
            let connection = TTLBackendConnectionSetup.create(credentials: credentials, device: device, backend: backend)
            self.device = device
            self.credentials = credentials
            self.backend = backend
            self.serviceClient.connect(toServer: connection)
        }
    }
    
    func onServerConnected(deviceIdentifier: String) {
        stateHandler.changeState(on: "onServerConnected", of: deviceIdentifier, from: [.connectingToBackend], to: .activating)
        {
            self.serviceClient.startActivation(forDevice: deviceIdentifier)
        }
    }
    
    func onServerDisconnected(deviceIdentifier: String, reason: ServerDisconnectReason) {
        var error = ActivationError.backendDisconnected
        if (reason == .wrongCredentials)
        {
            error = ActivationError.wrongCredentials
        }
        else if (reason == .wrongDevice)
        {
            error = ActivationError.wrongDevice
        }
        stateHandler.changeState(
            on: "onServerDisconnected",
            of: deviceIdentifier,
            from: [.activating, .connectingToBackend],
            to: .finished(error:error))
    }
    
    func onActivationFailed(_ deviceIdentifier: String, reason: TTLServiceActivationFailedReason) {
        stateHandler.changeState(
            on: "onActivationFailed",
            of: deviceIdentifier,
            from: [.activating],
            to: .finished(error:ActivationError.activationFailed))
        {
            self.serviceClient.disconnectFromServer()
        }
    }
    
    func onActivationSuccessful(_ deviceIdentifier: String) {
        stateHandler.changeState(on: "onActivationSuccessful", of: deviceIdentifier, from: [.activating], to: .finished(error:nil))
        {
            guard let credentials = self.credentials, let backend = self.backend else {return}
            self.serviceClient.disconnectFromServer()
            self.userService.credentials = credentials
            self.userService.deviceIdentifier = deviceIdentifier
            self.userService.backend = backend
        }
    }
    
    func onDeviceDisconnected(deviceIdentifier: String) {
        stateHandler.changeState(on: "onDeviceDisconnected", of: deviceIdentifier, from: [.connectingToBackend, .activating], to: .finished(error:.deviceDisconnected))
        {
            self.serviceClient.disconnectFromServer()
        }
    }
    
    func onStateChanged(newState: LoginActivationPresenterState) {
        delegate?.onActivationStateChange(state: newState)
    }
}

protocol LoginActivationPresenterDelegate : class {
    func onActivationStateChange(state:LoginActivationPresenterState)
}
