import UIKit

/**
 ViewController that handles the login sceen.
 */
class LoginViewController: UIViewController, UIPickerViewDataSource, UIPickerViewDelegate, UITextFieldDelegate, LoginDevicePresenterDelegate, LoginActivationPresenterDelegate {

    @IBOutlet weak var loginSectionView: UIStackView!
    @IBOutlet weak var activationSectionView: UIStackView!
    @IBOutlet weak var errorSectionView: UIView!
    @IBOutlet weak var userNameTextField: UITextField!
    @IBOutlet weak var passwordTextField: UITextField!
    @IBOutlet weak var productionSwitch: UISwitch!
    @IBOutlet weak var deviceTextField: UITextField!
    @IBOutlet weak var statusLabel: UILabel!
    @IBOutlet weak var errorLabel: UILabel!
    
    var devicePicker : UIPickerView!
    let loginDevicePresenter = LoginDevicePresenter(connectionService: ServiceFactory.connectionService)
    let activationPresenter = LoginActivationPresenter(serviceClient: ServiceFactory.serviceClient, connectionService: ServiceFactory.connectionService, userService: ServiceFactory.userService)
    
    override func viewDidLoad() {
        super.viewDidLoad()
        // Do any additional setup after loading the view, typically from a nib.
        devicePicker = UIPickerView()
        devicePicker.delegate = self
        devicePicker.dataSource = self
        deviceTextField.inputView = devicePicker
        loginDevicePresenter.delegate = self
        activationPresenter.delegate = self
    }

    func onDeviceListChanged() {
        devicePicker.reloadAllComponents()
        deviceTextField.text = loginDevicePresenter.deviceDisplayText()
    }
    
    /// Jumps to the next textfield after the return button was pressed on the keyboard.
    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        let nextTag = textField.tag + 1
        // Try to find next responder
        if let nextResponder = textField.superview?.viewWithTag(nextTag) {
            // Found next responder, so set it
            nextResponder.becomeFirstResponder()
        } else {
            // Not found, so remove keyboard
            textField.resignFirstResponder()
        }
        return false
    }
    
    /// Login button pressed. Start the activation process and proceed to main screen if activation successful.
    @IBAction func onLogin(_ sender: UIButton) {
        do
        {
            let userName = try userNameTextField.validatedText(validationType: .required)
            let password = try passwordTextField.validatedText(validationType: .required)
            let device = try deviceTextField.validatedText(validationType: .required)
            let production = productionSwitch.isOn
            
            loginSectionView.isUserInteractionEnabled = false
            activationSectionView.isHidden = false
            errorSectionView.isHidden = true
            activationPresenter.activate(userName: userName, password: password, device: device, prod: production)
        }
        catch let error as ValidatorError
        {
            error.source?.notifyError()
        }
        catch
        {
            // not happening
        }
    }
    
    /// Callback by the LoginActivationPresenter informing about the progress of the activation
    func onActivationStateChange(state: LoginActivationPresenterState) {
        switch state {
        case .finished(let error):
            activationSectionView.isHidden = true
            if error != nil {
                errorSectionView.isHidden = false
                errorLabel.text = state.toLocalizedString()
            } else {
                performSegue(withIdentifier: "showMainScreen", sender: self)
            }
            loginSectionView.isUserInteractionEnabled = true
        default:
            statusLabel.text = state.toLocalizedString()
        }
    }
    
    func numberOfComponents(in pickerView: UIPickerView) -> Int {
        return 1
    }
    
    func pickerView(_ pickerView: UIPickerView, numberOfRowsInComponent component: Int) -> Int {
        return loginDevicePresenter.availableDevices.count
    }
    
    func pickerView(_ pickerView: UIPickerView, titleForRow row: Int, forComponent component: Int) -> String? {
        return loginDevicePresenter.availableDevices[row]
    }
    
    func pickerView(_ pickerView: UIPickerView, didSelectRow row: Int, inComponent component: Int) {
        loginDevicePresenter.selectDevice(row)
        deviceTextField.text = loginDevicePresenter.deviceDisplayText()
        deviceTextField.resignFirstResponder()
    }
}
