import os

/**
 Presenter handling the firmware update process. During the firmware update process a new LINK100 firmware is downloaded from
 the server, uploaded to the device and finally installed on the device.
 
 The presenter uses a simplified state mashine to handle the activation process. The state mashine for the activation looks
 as follows, with states being enclosed in brackets ( [state] ) and events depicted as arrows ( -event-> ):
 
     // A new firmware update is available, wait for the user to trigger the update.
     [start] | [finish] -onFirmwareUpdateAvailable-> [updateAvailable]
     // Start the firmware update process, establish backend connection via serviceClient.connect(...)
     [updateAvailable] -startFirmwareUpdate-> [connectingToBackend]
     // Connection established, start the firmware update via serviceClient.updateFirmware(...)
     [connectingToBackend] -onServerConnected-> [downloading]
     // Download of the firmwareupdate is in progress.
     [downloading] -downloadStarted--> [uploading]
                   -downloadProgress->
                   -downloadFinished->
     // Uploading of the firmware to the device is in progress.
     [downloading] | [uploading] -uploadStarted--> [uploading]
                                 -uploadProgress->
                                 -uploadFinished->
     // Upload has finished firmware will be installed and device will reboot, it's save to disconnect from the backend now
     [uploading] -installRequested-> [rebooting]
     // The device reconnected, installation and reboot has finished.
     [rebooting] -onDeviceConnected-> [finished]
     // An error occured during firmware download, abort and disconnect from the backend.
     [downloading] -downloadFailed-> [finished]
     // An error occured during upload of the firmware, abort and disconnect from the backend.
     [downloading] | [uploading] -uploadFailed-> [finished]
 
 The presenter needs to handle additional error events. As the SDK works in a stateless fashion, it's not aware that an
 firmware update is in progress once it was triggered. If the device or server connection fails inbetween it will not send
 a dedicated firmware update fail event. However, it will inform about device or server disconnection via alternative service
 callbacks. Thus these are the additional error events the SDK needs to handle:
 
     // server connection lost during or before download, afterwards download a backend connection is not required anymore
     [connectingToBackend] | [downloading] -onServerDisconnected-> [finished]
     // The device connection was lost while uploading the firmware update.
     [uploading] -onDeviceDisconnected-> [finished]
 */
class FirmwareUpdatePresenter: NSObject , TTLServiceEventDelegate, ConnectionServiceDelegate, StateHandlerDelegate {
    private let serviceClient : TTLServiceClient
    private let userService: UserService
    private let connectionService: ConnectionService
    
    private(set) var stateHandler : StateHandler<FirmwareUpdatePresenter>
    private var firmwareUpdateParameter: TTLServiceFirmwareUpdateParameter?
    weak var delegate : FirmwareUpdatePresenterDelegate?
    
    init(serviceClient : TTLServiceClient, userService: UserService, connectionService: ConnectionService) {
        self.serviceClient = serviceClient
        self.userService = userService
        self.connectionService = connectionService
        self.stateHandler = StateHandler(initialState: .start, deviceIdentifier: userService.deviceIdentifier)
        super.init()
        self.stateHandler.delegate = self
        self.serviceClient.addServiceDelegate(self)
        self.connectionService.addDelegate(delegate: self)
    }
    
    var state: FirmwareUpdatePresenterState {
        return stateHandler.state
    }
    
    func startFirmwareUpdate() {
        stateHandler.changeState(on: "startFirmwareUpdate", from: [.updateAvailable], to: .connectingToBackend) {
            self.serviceClient.connect(toServer: TTLBackendConnectionSetup.create(
                credentials: self.userService.credentials, device: self.userService.deviceIdentifier, backend: self.userService.backend))
        }
    }
    
    //
    // MARK: - TTLServiceEventDelegate
    //
    
    /// new firmware update available (availability of a new firmware update will be checked by the SDK each time
    /// a device is connected and a server connection becomes available)
    func onFirmwareUpdateAvailable(_ deviceIdentifier: String, response: TTLServiceFirmwareUpdateAvailableResponse) {
        stateHandler.changeState(on: "onFirmwareUpdateAvailable", of: deviceIdentifier, from: [.finished(error: nil), .start, .updateAvailable], to: .updateAvailable){
            self.firmwareUpdateParameter = TTLServiceFirmwareUpdateParameter.init(response: response)
        }
    }
    
    /// main callback from the SDK informing about the progress of a firmware update
    func onFirmwareUpdateStatusUpdate(_ deviceIdentifier: String, response: TTLServiceFirmwareUpdateStatusResponse) {
        switch response.status {
        case .downloadStarted, .downloadProgress, .downloadFinished :
            stateHandler.changeState(on: "downloading", of: deviceIdentifier, from: [.downloading(0)],
                                     to: .downloading(response.progress))
        case .downloadFailed:
            stateHandler.changeState(on: "downloadFailed", of: deviceIdentifier, from: [.downloading(0)],
                                     to: .finished(error: .serverCommunicationFailed)) {
                self.serviceClient.disconnectFromServer()
            }
        case .uploadStarted, .uploadProgress, .uploadFinished:
            stateHandler.changeState(on: "uploading", of: deviceIdentifier, from: [.downloading(0), .uploading(0)],
                                     to: .uploading(response.progress))
        case .uploadFailed:
            stateHandler.changeState(on: "uploadFailed", of: deviceIdentifier, from: [.downloading(0), .uploading(0)],
                                     to: .finished(error: .deviceCommunicationFailed)) {
                self.serviceClient.disconnectFromServer()
            }
        case .installRequested:
            stateHandler.changeState(on: "installRequested", of: deviceIdentifier, from: [.uploading(0)],
                                     to: .rebooting) {
                self.serviceClient.disconnectFromServer()
            }
        }
    }
    
    //
    // MARK: - ConnectionServiceDelegate
    //
    
    func onServerConnected(deviceIdentifier: String) {
        stateHandler.changeState(on: "onServerConnected", of: deviceIdentifier, from: [.connectingToBackend], to: .downloading(0)) {
            self.serviceClient.updateFirmware(self.firmwareUpdateParameter!, deviceIdentifier: deviceIdentifier)
        }
    }
    
    func onServerDisconnected(deviceIdentifier: String, reason: ServerDisconnectReason) {
        stateHandler.changeState(on: "onServerDisconnected", of: deviceIdentifier, from: [.connectingToBackend, .downloading(0)], to: .finished(error: .serverCommunicationFailed))
    }
    
    func onDeviceConnected(deviceIdentifier: String) {
        stateHandler.changeState(on: "onDeviceConnected", of: deviceIdentifier, from: [.rebooting], to: .finished(error: nil))
    }
    
    func onDeviceDisconnected(deviceIdentifier: String) {
        stateHandler.changeState(on: "onDeviceDisconnected", of: deviceIdentifier, from: [.uploading(0)], to: .finished(error: .deviceCommunicationFailed)) {
            self.serviceClient.disconnectFromServer()
        }
    }
    
    //
    // MARK: - StateHandlerDelegate
    //
    
    func onStateChanged(newState: FirmwareUpdatePresenterState) {
        delegate?.onFirmwareUpdateStateChange()
    }
}

protocol FirmwareUpdatePresenterDelegate : class {
    func onFirmwareUpdateStateChange()
}
