import CoreFoundation
import os

/**
 Presenter requesting and preparing real time data from a LINK100 device for displaying.
 */
class RealTimeDataPresenter : NSObject, TTLServiceODBDataSubscriptionProtocol {
    private let serviceClient : TTLServiceClient
    private let userService : UserService
    private var lastDatagram : TTLOBDDatagram? = nil
    private(set) var enabled = false
    weak var delegate : RealTimeDataPresenterDelegate? = nil
    
    init(serviceClient : TTLServiceClient, userService: UserService) {
        self.serviceClient = serviceClient
        self.userService = userService
        super.init()
    }
    
    /// Subscribe for realtime data, neccessary only once even if the device disconnects inbetween.
    /// Data will be provided instantly after this call if a device is connected.
    func startRealTimeData() {
        enabled = true
        serviceClient.subscribe(toOBDData: self, deviceIdentifier: userService.deviceIdentifier)
    }
    
    /// Unsubscribe from realtime data, neccessary only once even if the device is disconnected.
    /// Real time data transmission will end after this call.
    func stopRealTimeData() {
        enabled = false
        serviceClient.unsubscribe(fromOBDData: self, deviceIdentifier: userService.deviceIdentifier)
    }
    
    /// Callback by the SDK delivering the real time data.
    func onOBDDatagram(_ datagram: TTLOBDDatagram) {
        self.lastDatagram = datagram
        DispatchQueue.main.async {
            self.delegate?.onDatagramChanged()
        }
    }
    
    /// Last odometer value in km.
    var odometer: String {
        get {
            guard let odometer = lastDatagram?.odometer?.doubleValue else
            {
                return "-"
            }
            return localize(key: "Unit.Kilometer", withArgs: Double(odometer / 1000))
        }
    }
    
    /// Last rpm value.
    var rpm: String {
        get {
            guard let rpm = lastDatagram?.engineRpm?.doubleValue else
            {
                return "-"
            }
            return localize(key: "Unit.Rpm", withArgs: rpm)
        }
    }
}

protocol RealTimeDataPresenterDelegate : class {
    func onDatagramChanged()
}
