import os

/**
 Presenter handling the synchronization process on the main screen. The synchronization is the process which uploads
 LINK100 device trip data to the TomTom backend.
 
 The presenter uses a simplified state mashine to handle the synchronization process. The main state mashine looks
 as follows, with states being enclosed in brackets ( [state] ) and events depicted as arrows ( -event-> ):
 
     // start of the sync process, establish backend connection via serviceClient.connect(...)
     [start] | [finished] -startSync-> [connectingToBackend]
     // backend connection established, trigger synchronization via serviceClient.startSync(...)
     [connectingToBackend] -onServerConnected-> [connectingToBackend]
     // downloading of sync data started
     [connectingToBackend] -downloadStarted-> [downloading]
      // downloading sync data
     [downloading] -downloadProgress-> [downloading]
      // dowloading of sync data finished, wait for uploading of sync data
     [downloading] -downloadSuccessful-> [uploading]
     // sync finished, disconnect from backend if neccessary via serviceClient.disconnectFromServer(...)
     [uploading] -uploadSuccessful-> [finished]
     // general sync error during download or upload of sync data, sync failed
     [connectingToBackend] [clockSynchronization] [downloading] -downloadFailed-> [finished]
 
 Additionally the presenter needs to handle the case when the LINK device was unplugged since the last sync attempt.
 In this case the device needs to resynchronize it's internal clock with the backend. That's why this process is referred
 to as clock sync by the SDK. Since it's a potential attempt by the user to veil his/her recent driving the SDK informs
 about such events and allows the client application to request a reason from the user why the device was unplugged.
 These are the additional state changes neccessary to handle the clock sync process.
 
     // the device was unplugged, a reason for the disconnect needs to be provided
     [connectingToBackend] -onClockSyncReasonWhileDownloadRequired-> [clockSynchronization]
     // notify the SDK about the reason for unplugging the device, via serviceClient.sendClockSyncReason(...)
     [clockSynchronization] -sendClockSyncReason-> [clockSynchronization]
     // clocksynchronization finished, sync data download starts
     [clockSynchronization] -downloadStarted-> [downloading]
     // clock sync faild
     [connectingToBackend] -onClockSyncFailed-> [finished]
 
 Appart from the previous events the presenter needs to handle some additional error events.
 As the SDK works in a stateless fashion, it's not aware that an synchronization is in progress once it was triggered.
 If the device or server connection fails inbetween it will not fail with a dedicated sync fail event. However, it will
 inform about device or server disconnection via alternative service callbacks. Thus these are the additional error
 events the presenter needs to handle:
 
     // server connection lost during sync
     [connectingToBackend] | [clockSynchronization] | [downloading] | [uploading] -onServerDisconnected-> [finished]
     // The device connection was lost while synchronizing. After downloading a device connetion is not required anymore
     [start] | [connectingToBackend] | [clockSynchronization] | [downloading] -onDeviceDisconnected-> [finished]
 
 We included an additional edge case that usually needs no consideration, but is mentioned here for completeness.
 Before a firmware update is installed the LINK device will attempt to sync all it's data. Since sync and firmware update
 process are handled in the same controller we included an additional state transition in this presenter to show the sync
 during firmware update as well. Note that this is not required as the firmware update will proceed no matter if the sync
 failed or not.
 
     // sync during firmware update started
     [start] | [finished] -downloadStarted-> [downloading]
 */
class SynchronizationPresenter: NSObject, TTLServiceEventDelegate, ConnectionServiceDelegate, StateHandlerDelegate {
    private let serviceClient: TTLServiceClient
    private let userService: UserService
    private let connectionService: ConnectionService
    
    /// handler that takes care of state transistions during sync process
    private var stateHandler : StateHandler<SynchronizationPresenter>
    /// total amount of bytes to download during sync
    private var bytesToDownload = 0.0
    /// current amount of bytes downloaded during sync
    private var bytesDownloaded = 0.0
    /// delegate that is informed about events of the presenter
    weak var delegate : SynchronizationPresenterDelegate?
    
    init(serviceClient : TTLServiceClient, userService : UserService, connectionService: ConnectionService) {
        self.serviceClient = serviceClient
        self.userService = userService
        self.connectionService = connectionService
        self.stateHandler = StateHandler(initialState: .start, deviceIdentifier: userService.deviceIdentifier)
        super.init()
        self.stateHandler.delegate = self
        self.serviceClient.addServiceDelegate(self)
        self.connectionService.addDelegate(delegate: self)
    }
    
    func startSync(){
        // check that a device is actually connected, as the SDK will skip sync silently otherwise
        guard connectionService.connected(to: userService.deviceIdentifier) else
        {
            stateHandler.changeState(on: "onDeviceDisconnected", from: [.start, .finished(error: nil)], to: .finished(error: .deviceCommunicationFailed))
            return
        }
        // start the sync process
        stateHandler.changeState(on: "startSync", from: [.start, .finished(error: nil)], to: .connectingToBackend)
        {

            self.serviceClient.connect(toServer: TTLBackendConnectionSetup.create(
                credentials: self.userService.credentials,
                device: self.userService.deviceIdentifier,
                backend: self.userService.backend)
            )
        }
    }
    
    //
    // MARK: - TTLServiceEventDelegate
    //
    
    func sendClockSyncReason(_ reason: ClockSyncReason) {
        stateHandler.changeState(on: "sendClockSyncReason", from: [.clockSynchronization], to: .clockSynchronization) {
            self.serviceClient.sendClockSyncReason(
                forDownload: TTLServiceDeviceUnplugReason(reason: reason.value),
                deviceIdentifier: self.userService.deviceIdentifier)
        }
    }
    
    /// The device was unplugged since the last sync. A reason is required before continuing the sync process
    func onClockSyncReasonWhileDownloadRequired(_ deviceIdentifier: String) {
        if stateHandler.changeState(on: "onClockSyncReasonWhileDownloadRequired", of: deviceIdentifier, from: [.connectingToBackend], to: .clockSynchronization) {
            self.delegate?.onClockSyncReasonRequired()
        }
    }
    
    func onClockSyncFailed(_ deviceIdentifier: String, reason: TTLServiceClockSyncFailedReason) {
        stateHandler.changeState(on: "onClockSyncFailed", of: deviceIdentifier, from: [.clockSynchronization], to: .finished(error: .serverCommunicationFailed)) {
            self.serviceClient.disconnectFromServer()
        }
    }
    
    /// main callback from the SDK for receiving sync progress information
    func onDownloadStatusUpdate(_ deviceIdentifier: String, response downloadStateResponse: TTLServiceDownloadStatusResponse) {
        switch downloadStateResponse.progressState
        {
        case .downloadStarted:
            stateHandler.changeState(on: "downloadStarted", of: deviceIdentifier, from: [.start, .finished(error: nil), .connectingToBackend, .clockSynchronization], to: .downloading(0))
            {
                self.bytesToDownload = downloadStateResponse.progress!.doubleValue
                self.bytesDownloaded = 0
            }
        case .downloadProgress:
            let totalBytesDownloaded = bytesDownloaded + (downloadStateResponse.progress?.doubleValue ?? 0)
            let progress = totalBytesDownloaded / bytesToDownload * 100
            let progressRounded = Int(progress.rounded(FloatingPointRoundingRule.up))
            stateHandler.changeState(on: "downloadProgress", of: deviceIdentifier, from: [.downloading(0)], to: .downloading(progressRounded))
            {
                self.bytesDownloaded = totalBytesDownloaded
            }
        case .downloadSuccessful:
            stateHandler.changeState(on: "downloadSuccessful", of: deviceIdentifier, from: [.downloading(0)], to: .uploading)
        case .uploadSuccessful:
            stateHandler.changeState(on: "uploadSuccessful", of: deviceIdentifier, from: [.uploading], to: .finished(error: nil))
            {
                self.serviceClient.disconnectFromServer()
            }
        case.downloadFailed:
            stateHandler.changeState(on: "downloadFailed", of: deviceIdentifier, from: [.connectingToBackend, .clockSynchronization, .downloading(0)], to: .finished(error: .deviceCommunicationFailed))
            {
                self.serviceClient.disconnectFromServer()
            }
        }
    }
    
    //
    // MARK: - ConnectionServiceDelegate
    //
    
    func onDeviceDisconnected(deviceIdentifier: String) {
        stateHandler.changeState(
            on: "onDeviceDisconnected",
            of: deviceIdentifier,
            from: [.connectingToBackend, .clockSynchronization, .downloading(0)],
            to: .finished(error: .deviceCommunicationFailed))
        {
            self.serviceClient.disconnectFromServer()
        }
    }
    
    func onServerConnected(deviceIdentifier: String) {
        stateHandler.changeState(on: "onServerConnected", of: deviceIdentifier, from: [.connectingToBackend], to: .connectingToBackend)
        {
            self.serviceClient.startSync(forDevice: self.userService.deviceIdentifier)
        }
    }
    
    func onServerDisconnected(deviceIdentifier: String, reason: ServerDisconnectReason) {
        stateHandler.changeState(
            on: "onServerDisconnected",
            of: deviceIdentifier,
            from: [.connectingToBackend, .clockSynchronization, .downloading(0), .uploading],
            to: .finished(error: .serverCommunicationFailed))
    }
    
    //
    // MARK: - StateHandlerDelegate
    //
    
    func onStateChanged(newState: SynchronizationPresenterState) {
        delegate?.onSyncStateChange(state: newState)
    }
}

protocol SynchronizationPresenterDelegate : class {
    func onSyncStateChange(state: SynchronizationPresenterState)
    func onClockSyncReasonRequired()
}
