import Foundation

/**
 States the SynchronizationPresenter passes while running the synchronization process.
 */
enum SynchronizationPresenterState : Equatable, CustomDebugStringConvertible {
    /// Waiting for sychronization to be triggered.
    case start
    /// Synchronization started, connecting to backend.
    case connectingToBackend
    /// Device was unplugged since last sync, sychronizing clock with backend.
    case clockSynchronization
    /// Downloading sync data from the LINK device.
    case downloading(_ progress: Int)
    /// Uploading sync data to the backend.
    case uploading
    /// Synchronization finished successfully if `error` was `nil`, otherwise synchronization failed.
    case finished(error: SynchronizationError?)
    
    private var value: Int {
        switch self {
        case .start: return 0
        case .connectingToBackend: return 1
        case .clockSynchronization: return 2
        case .downloading: return 3
        case .uploading: return 4
        case .finished: return 5
        }
    }
    
    static func ==(lhs: SynchronizationPresenterState, rhs: SynchronizationPresenterState) -> Bool {
        // only consider the actual state, ignore concret error and progress values
        return lhs.value == rhs.value
    }
    
    var debugDescription: String {
        switch self {
        case .start: return "start"
        case .connectingToBackend: return "connectingToBackend"
        case .clockSynchronization: return "clockSynchronization"
        case .downloading(let progress): return String(format: "downloading %i", progress)
        case .uploading: return "uploading"
        case .finished(let error): return "finished (error: " + String(reflecting: error) + ")"
        }
    }
}
